@@ -1,7 +1,3 @@
-/*
- * $Id: wgdiplus.cpp,v 1.59.2.8 2017/08/29 11:45:16 markisch Exp $
- */
-
 /*
 Copyright (c) 2011-2017 Bastian Maerkisch. All rights reserved.
 
@@ -30,10 +26,10 @@ OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 // include iostream / cstdio _before_ syscfg.h in order
 // to avoid re-definition by wtext.h/winmain.c routines
 #include <iostream>
+#include <vector>
 extern "C" {
 # include "syscfg.h"
 }
-#define WINDOWS_LEAN_AND_MEAN
 #include <windows.h>
 #include <windowsx.h>
 #define GDIPVER 0x0110
@@ -57,7 +53,6 @@ using namespace Gdiplus;
 static bool gdiplusInitialized = false;
 static ULONG_PTR gdiplusToken;
 
-#define GWOPMAX 4096
 #define MINMAX(a,val,b) (((val) <= (a)) ? (a) : ((val) <= (b) ? (val) : (b)))
 const int pattern_num = 8;
 
@@ -235,7 +230,6 @@ SetFont_gdiplus(Graphics &graphics, LPRECT rect, LPGW lpgw, LPTSTR fontname, int
 		fontname = lpgw->deffontname;
 	if (size == 0)
 		size = lpgw->deffontsize;
-	size *= lpgw->fontscale;
 
 	/* make a local copy */
 	fontname = _tcsdup(fontname);
@@ -244,6 +238,9 @@ SetFont_gdiplus(Graphics &graphics, LPRECT rect, LPGW lpgw, LPTSTR fontname, int
 	_tcscpy(lpgw->fontname, fontname);
 	lpgw->fontsize = size;
 
+	// apply fontscale
+	size *= lpgw->fontscale;
+
 	/* set up font style */
 	INT fontStyle = FontStyleRegular;
 	LPTSTR italic, bold, underline, strikeout;
@@ -342,7 +339,7 @@ EnhancedSetFont()
 static unsigned
 EnhancedTextLength(char * text)
 {
-	LPWSTR textw = UnicodeText(enhanced_text, enhstate.lpgw->encoding);
+	LPWSTR textw = UnicodeTextWithEscapes(enhanced_text, enhstate.lpgw->encoding);
 	RectF box;
 	enhstate_gdiplus.graphics->MeasureString(textw, -1, enhstate_gdiplus.font, PointF(0, 0), enhstate_gdiplus.stringformat, &box);
 	free(textw);
@@ -353,7 +350,7 @@ EnhancedTextLength(char * text)
 static void
 EnhancedPutText(int x, int y, char * text)
 {
-	LPWSTR textw = UnicodeText(text, enhstate.lpgw->encoding);
+	LPWSTR textw = UnicodeTextWithEscapes(text, enhstate.lpgw->encoding);
 	Graphics *g = enhstate_gdiplus.graphics;
 	if (enhstate.lpgw->angle == 0) {
 		PointF pointF(x, y + enhstate.lpgw->tmDescent);
@@ -486,9 +483,9 @@ do_draw_gdiplus(LPGW lpgw, Graphics &graphics, LPRECT rect, enum draw_target tar
 	/* polylines and polygons */
 	int polymax = 200;			/* size of ppt */
 	int polyi = 0;				/* number of points in ppt */
-	PointF * ppt;				/* storage of polyline/polygon-points */
+	std::vector<PointF> ppt(polymax);	/* storage of polyline/polygon-points */
 	int last_polyi = 0;			/* number of points in last_poly */
-	PointF * last_poly = NULL;	/* storage of last filled polygon */
+	std::vector<PointF> last_poly(0);	/* storage of last filled polygon */
 	unsigned int lastop = -1;	/* used for plotting last point on a line */
 
 	/* filled polygons and boxes */
@@ -590,8 +587,6 @@ do_draw_gdiplus(LPGW lpgw, Graphics &graphics, LPRECT rect, enum draw_target tar
 					DashCapFlat);
 	solid_pen.SetLineJoin(lpgw->rounded ? LineJoinRound : LineJoinMiter);
 
-	ppt = (PointF *) malloc((polymax + 1) * sizeof(PointF));
-
 	htic = (lpgw->org_pointsize * MulDiv(lpgw->htic, rr - rl, lpgw->xmax) + 1);
 	vtic = (lpgw->org_pointsize * MulDiv(lpgw->vtic, rb - rt, lpgw->ymax) + 1);
 
@@ -647,7 +642,7 @@ do_draw_gdiplus(LPGW lpgw, Graphics &graphics, LPRECT rect, enum draw_target tar
 		}
 
 		/* finish last filled polygon */
-		if ((last_poly != NULL) &&
+		if ((!last_poly.empty()) &&
 			(((lastop == W_filled_polygon_draw) && (curptr->op != W_fillstyle)) ||
 			 ((curptr->op == W_fillstyle) && (curptr->x != unsigned(last_fillstyle))))) {
 			if (poly_graphics == NULL) {
@@ -655,14 +650,13 @@ do_draw_gdiplus(LPGW lpgw, Graphics &graphics, LPRECT rect, enum draw_target tar
 				SmoothingMode mode = graphics.GetSmoothingMode();
 				if (lpgw->antialiasing && !lpgw->polyaa)
 					graphics.SetSmoothingMode(SmoothingModeNone);
-				graphics.FillPolygon(fill_brush, last_poly, last_polyi);
+				graphics.FillPolygon(fill_brush, last_poly.data(), last_polyi);
 				graphics.SetSmoothingMode(mode);
 			} else {
-				poly_graphics->FillPolygon(fill_brush, last_poly, last_polyi);
+				poly_graphics->FillPolygon(fill_brush, last_poly.data(), last_polyi);
 			}
 			last_polyi = 0;
-			free(last_poly);
-			last_poly = NULL;
+			last_poly.clear();
 		}
 
 		/* handle layer commands first */
@@ -1337,7 +1331,7 @@ do_draw_gdiplus(LPGW lpgw, Graphics &graphics, LPRECT rect, enum draw_target tar
 			/* a point of the polygon is coming */
 			if (polyi >= polymax) {
 				polymax += 200;
-				ppt = (PointF *) realloc(ppt, (polymax + 1) * sizeof(PointF));
+				ppt.reserve(polymax + 1);
 			}
 			ppt[polyi].X = xdash;
 			ppt[polyi].Y = ydash;
@@ -1351,7 +1345,7 @@ do_draw_gdiplus(LPGW lpgw, Graphics &graphics, LPRECT rect, enum draw_target tar
 			//bool same_rot = true;
 
 			// Test if successive polygons share a common edge:
-			if ((last_poly != NULL) && (polyi > 2)) {
+			if ((!last_poly.empty()) && (polyi > 2)) {
 				// Check for a common edge with previous filled polygon.
 				for (i = 0; (i < polyi) && !found; i++) {
 					for (k = 0; (k < last_polyi) && !found; k++) {
@@ -1379,7 +1373,7 @@ do_draw_gdiplus(LPGW lpgw, Graphics &graphics, LPRECT rect, enum draw_target tar
 
 				int extra = polyi - 2;
 				// extend buffer to make room for extra points
-				last_poly = (PointF *) realloc(last_poly, (last_polyi + extra + 1) * sizeof(PointF));
+				last_poly.reserve(last_polyi + extra + 1);
 				/* TODO: should use memmove instead */
 				for (int n = last_polyi - 1; n >= k; n--) {
 					last_poly[n + extra].X = last_poly[n].X;
@@ -1392,22 +1386,20 @@ do_draw_gdiplus(LPGW lpgw, Graphics &graphics, LPRECT rect, enum draw_target tar
 				}
 				last_polyi += extra;
 			} else {
-				if (last_poly != NULL) {
+				if (!last_poly.empty()) {
 					if (poly_graphics == NULL) {
 						// changing smoothing mode is still necessary in case of new/unknown code paths
 						SmoothingMode mode = graphics.GetSmoothingMode();
 						if (lpgw->antialiasing && !lpgw->polyaa)
 							graphics.SetSmoothingMode(SmoothingModeNone);
-						graphics.FillPolygon(fill_brush, last_poly, last_polyi);
+						graphics.FillPolygon(fill_brush, last_poly.data(), last_polyi);
 						graphics.SetSmoothingMode(mode);
 					} else {
-						poly_graphics->FillPolygon(fill_brush, last_poly, last_polyi);
+						poly_graphics->FillPolygon(fill_brush, last_poly.data(), last_polyi);
 					}
-					free(last_poly);
 				}
 				// save the current polygon
-				last_poly = (PointF *) malloc(sizeof(PointF) * (polyi + 1));
-				memcpy(last_poly, ppt, sizeof(PointF) * (polyi + 1));
+				last_poly = ppt;
 				last_polyi = polyi;
 			}
 
@@ -1653,7 +1645,7 @@ do_draw_gdiplus(LPGW lpgw, Graphics &graphics, LPRECT rect, enum draw_target tar
 		delete cb;
 	if (font)
 		delete font;
-	free(ppt);
+	ppt.clear();
 }
 
 
@@ -1745,7 +1737,11 @@ SaveAsBitmap(LPGW lpgw)
 	Ofn.lpstrDefExt = L"png";
 
 	if (GetSaveFileNameW(&Ofn) != 0) {
-		Bitmap * bitmap = Bitmap::FromHBITMAP(lpgw->hBitmap, 0);
+		// Note that there might be a copy in lpgw->hBitmap., but documentation
+		// says we may not use that (although it seems to work).
+		// So we get a new copy of the screen:
+		HBITMAP hBitmap = GraphGetBitmap(lpgw);
+		Bitmap * bitmap = Bitmap::FromHBITMAP(hBitmap, 0);
 		UINT ntype = Ofn.nFilterIndex - 1;
 #if 0
 		LPWSTR wtype = pImageCodecInfo[ntype].FormatDescription;
@@ -1757,6 +1753,7 @@ SaveAsBitmap(LPGW lpgw)
 #endif
 		bitmap->Save(Ofn.lpstrFile, &(pImageCodecInfo[ntype].Clsid), NULL);
 		delete bitmap;
+		DeleteObject(hBitmap);
 	}
 	free(filter);
 }
